#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Shairport Sync Metadata Plugin
# Developed by Jeremy Swancoat

import os
import sys
import time
import socket
import requests
from threading import Thread
from shairport import Shairport


# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process.

##################################################################################################
class Plugin(indigo.PluginBase):
	##############################################################################################
	# Required Methods
	##############################################################################################
	##############################################################################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
		# Call the base class's init method
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		# Setting debug to True will show you verbose debugging information in the Indigo Event Log
		# LOTS of debugLog commands are placed throughout this routine.
		if 'showDebugInfo' in pluginPrefs:
			self.debug = pluginPrefs['showDebugInfo']
		else:
			self.debug = False
		self.devlist = {}
		self.shcom = {}
		
		
	########################################
	def __del__(self):
		indigo.PluginBase.__del__(self)

	##############################################################################################
	# Non-Required Methods (but still defined by Indigo)
	##############################################################################################

	########################################
	def startup(self):
		# no real startup process. All procedures are basically run at the device level.
		self.logger.debug(u"Startup Called")

	def shutdown(self):
		# not much to shut down after devices are shut down.
		self.logger.debug(u"Shutdown Called")

	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		# Validate that appropriate selections were made in device creation / modification dialogs.
		# Ensure that a device isn't already set up for this Shairport Sync
		errorDict = indigo.Dict()
		self.logger.debug("Validating Device")
		self.logger.debug("IP Address: " + valuesDict["ipaddress"])
		### Ensure IP Address is valid
		if valuesDict["ipaddress"].count('.') == 3:
			pass
		else:
			errorDict["ipaddress"] = "Invalid IP Address."

		### Logic to see if device is in use
		### If this is just a modification to the current device, don't check its ID against the currently used
		if devId == 0:
			if valuesDict["ipaddress"] in self.devlist:
				self.logger.debug("IP Address Already Assigned to Device")
				errorDict["ipaddress"] = "Already in use"
			else:
				self.logger.debug("IP Address Avaialble")
	

		if len(errorDict) > 0:
			return (False, valuesDict, errorsDict)
		return (True, valuesDict)
		
	def loadIP(self, valuesDict, typeId, devId):
		#Just gets the local machine's IP address for the config dialog.
		hostip = socket.gethostbyname(socket.gethostname())
		self.logger.debug("Host IP Address: " + hostip)
		valuesDict["ipaddress"] = hostip
		return valuesDict

	def deviceStartComm(self, dev):
		self.devlist[dev.pluginProps["ipaddress"]] = dev
		self.logger.debug("deviceStartComm called for " + dev.name)
		self.logger.debug("Just added: " + dev.pluginProps["ipaddress"] + " to devlist")
		# create a new Shairport object and start communication.
		# create the object in shcom dictionary and use the device id as the key
		self.shcom[dev.id] = Shairport(dev, self.logger, indigo.server.broadcastToSubscribers)
		#self.shcom[dev.id].startSocket()
		# if broadcasting metadata, send the broadcast started message for this device
		if dev.pluginProps["shareMetadata"] == True:
			indigo.server.broadcastToSubscribers(u"startingBcast")
			self.logger.info("Broadcast Started for " + dev.name)		
		# Start runDevConcurrentThread... as a Thread.
		Thread(target=self.shcom[dev.id].runDevConcurrentThread).start()

	def deviceStopComm(self, dev):
		self.logger.debug("deviceStopComm called for " + dev.name)
		self.shcom[dev.id].stopDevConcurrentThread()
		#self.shcom[dev.id].stopSocket()
		# if broadcasting metadata, send the broadcast ended message for this device
		if dev.pluginProps["shareMetadata"] == True:
			indigo.server.broadcastToSubscribers(u"endingBcast")
			self.logger.info("Broadcast Ended for " + dev.name)
		# remove from device list
		del self.devlist[dev.pluginProps["ipaddress"]]
		self.logger.debug("Just deleted: " + dev.pluginProps["ipaddress"] + " from devlist")

	def validatePrefsConfigUi(self, valuesDict):
		errorsDict = indigo.Dict()
		if valuesDict[u'showDebugInfo'] == True:
			self.debug = True
		else:
			self.debug = False

		if len(errorsDict) > 0:
			return (False, valuesDict, errorsDict)
		return (True, valuesDict)

	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		self.logger.debug("closedPrefsConfigUI() called")
		
	def commandToSource(self, pluginAction, dev):
		self.logger.debug("Skip Forward " + dev.name)
		clientip = ""
		artoken = ""
		dacpid =""
		srcremport = ""
		try:
			clientip = self.shcom[dev.id].statedict["clientip"]
			artoken = self.shcom[dev.id].statedict["artoken"]		
			dacpid = self.shcom[dev.id].statedict["dacpid"]
			srcremport = self.shcom[dev.id].statedict["srcremport"]
		except:
			pass
			
		if srcremport == "":
			self.logger.info("DACP Remote Port not available for " + dev.name)
		elif clientip == "":
			self.logger.info("No client IP available for device streaming to " + dev.name)
			self.logger.info("Try disconnecting and reconnection client from Shairport Sync")
		else:
			self.logger.debug(dev.name + " IP Address: " + clientip)
			self.logger.debug(dev.name + " Port: " + srcremport)
			self.logger.debug(dev.name + " DACP ID: " + dacpid)
			self.logger.debug(dev.name + " Active Remote Token: " + artoken)
		
			# Get ipv4 address for any ipv6 addresses.
			if ":" in clientip:
				triple = socket.gethostbyaddr(clientip)
				clientip = socket.gethostbyname(triple[0])
		
			url = "http://" + clientip + ":" + srcremport +"/ctrl-int/1/"+pluginAction.props.get("selCommand")
			self.logger.debug(url)
		
			parameters = {'Active-Remote': artoken}
			req = requests.get(url,  headers = parameters)
			req = None