#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Shairport Sync Metadata Plugin
# Developed by Jeremy Swancoat

import os
import sys
import time
import datetime
import socket
import struct


# Note the "indigo" module is automatically imported and made available inside
# our global name space by the host process.

##################################################################################################
class Shairport(object):

	def  __init__(self, dev, logger, broadcastToSubscribers):
		
		self.dev = dev
		self.logger = logger
		self.stopThread = False
		self.broadcastToSubscribers = broadcastToSubscribers
		self.statedict = {}
		
		self.logger.debug("Shairport instance initiated for " + self.dev.name)

		# Dictionary key is the metadata code (ssnc is from Shairplay; core is from iTunes or whatever)
		# Value is a tuple including description, data type, device state it's used to adjust and booleans indicating if the 
		# data should update a device state, be kept as a local variable, but not device state (stuff like DACP-ID - no one
		# needs to see it... be we need to keep it) and if it should be broadcast

		self.coderef = {'ssncPICT': ("Picture", 9, "unused", False, False, False)}
		self.coderef['ssncpfls'] = ("Play Stream Flush", 0, "nowplaying", True, True, True)
		self.coderef['ssncpend'] = ("Play Stream End", 0, "nowplaying", True, True, True)
		self.coderef['ssncclip'] = ("Client IP", 9, "clientip", False, True, True)
		self.coderef['ssncsvip'] = ("Server IP", 9, "serverip", False, False, False)
		self.coderef['ssncsnua'] = ("User Agent", 9, "useragent", False, False, True)
		self.coderef['ssncacre'] = ("Active Remote Token", 9, "artoken", False, True, True)
		self.coderef['ssncdaid'] = ("DACP-ID", 9, "dacpid", False, True, True)
		self.coderef['ssncpbeg'] = ("Play Stream Begin", 0, "nowplaying", True, True, True)
		self.coderef['ssncpvol'] = ("Play Stream Volume", 9, "unused", False, False, False)
		self.coderef['ssncmdst'] = ("Starting Metadata", 2, "metadata", False, False, False)
		self.coderef['ssncprsm'] = ("Play Stream Resume", 0, "nowplaying", True, True, True)
		self.coderef['ssncprgr'] = ("Progress", 14, "tracktime", True, False, True)
		self.coderef['ssncmden'] = ("Metadata Ended", 2, "metadata", False, False, False)
		self.coderef['ssncpcst'] = ("Starting Picture", 13, "unused", False, False, False)
		self.coderef['ssncpcen'] = ("Picture Ended", 13, "unused", False, False, False)
		self.coderef['ssncsnam'] = ("Device Name", 9, "devicename", True, False, True)
		self.coderef['ssncdapo'] = ("Source Remote Port", 9, "srcremport", False, True, True)
		self.coderef['coremikd'] = ('Item Kind', 2, "unused", False, False, False)
		self.coderef['coreasal'] = ("Album", 9, "album", True, False, True)
		self.coderef['coreasar'] = ("Artist", 9, "artist", True, False, True)
		self.coderef['coreasbr'] = ("Bitrate", 3, "bitrate", True, False, True)
		self.coderef['coreascm'] = ("Comment", 9, "comment", True, False, True)
		self.coderef['coreasco'] = ("Compilation", 1, "unused", False, False, False)
		self.coderef['coreascp'] = ("Composer", 9, "composer", True, False, True)
		self.coderef['coreasda'] = ("Date Added", 10, "added", True, False, True)
		self.coderef['coreasdm'] = ("Date Modified", 10, "modified", True, False, True)
		self.coderef['coreasdc'] = ("Disc Count", 3, "unused", True, False, False)
		self.coderef['coreasdn'] = ("Disc Number", 3, "unused", True, False, False)
		self.coderef['coreaseq'] = ("EQ Preset", 9, "unused", False, False, False)
		self.coderef['coreasgn'] = ("Genre", 9, "genre", True, False, True)
		self.coderef['coreasdt'] = ("Description", 9, "filedesc", True, False, True)
		self.coderef['coreasrv'] = ("Relative Volume", 2, "unused", False, False, False)
		self.coderef['coreassr'] = ("Sample Rate", 5, "samplerate", True, False, True)
		self.coderef['coreassz'] = ("Size", 5, "unused", False, False, False)
		self.coderef['coreasst'] = ("Start Time", 5, "unused", False, False, False)
		self.coderef['coreassp'] = ("Stop Time", 5, "unused", False, False, False)
		self.coderef['coreastm'] = ("Time", 5, "unused", False, False, False)
		self.coderef['coreastc'] = ("Track Count", 3, "trackcount", True, False, True)
		self.coderef['coreastn'] = ("Track Number", 3, "tracknumber", True, False, True)
		self.coderef['coreasur'] = ("User Rating", 1, "unused", False, False, False)
		self.coderef['coreasyr'] = ("Year", 3, "year", True, False, True)
		self.coderef['coreasfm'] = ("Format", 9, "format", True, False, True)
		self.coderef['coreminm'] = ("Item Name", 9, "name", True, False, True)
		self.coderef['coremper'] = ("Persistent ID", 7, "unused", False, False, False)
		self.coderef['coreasdb'] = ("Disabled", 1, "unused", False, False, False)
		self.coderef['coreaeNV'] = ("Norm Volume", 5, "unused", False, False, False)
		self.coderef['coreasdk'] = ("Data Kind", 1, "unused", False, False, False)
		self.coderef['coreasbt'] = ("BPM", 3, "bpm", True, False, True)
		self.coderef['coreagrp'] = ("Grouping", 9, "unused", False, False, False)
		self.coderef['coreascd'] = ("Codec Type", 5, "unused", False, False, False)
		self.coderef['coreascs'] = ("Codec Subtype", 5, "unused", False, False, False)
		self.coderef['coreaePC'] = ("Is Podcast", 1, "podcast", True, False, True)
		self.coderef['coreasct'] = ("Category", 9, "unused", False, False, False)
		self.coderef['coreascn'] = ("Content Description", 9, "unused", False, False, False)
		self.coderef['coreascr'] = ("Content Rating", 1, "unused", False, False, False)
		self.coderef['coreaeHV'] = ("Has Video", 1, "unused", False, False, False)
		self.coderef['coreaeMK'] = ("Media Kind", 1, "unused", False, False, False)
		self.coderef['coreaeSN'] = ("Series Name", 9, "unused", False, False, False)
		self.coderef['coreaeEN'] = ("Episode Number", 9, "unused", False, False, False)
		self.coderef['coreaeES'] = ("Episode Sort", 5, "unused", False, False, False)
		self.coderef['coreaeSU'] = ("Season Number", 5, "unused", False, False, False)
		self.coderef['coreaeGH'] = ("Gapless Heur", 5, "unused", False, False, False)
		self.coderef['coreaeGD'] = ("Gapless Enc Dr", 5, "unused", False, False, False)
		self.coderef['coreaeGU'] = ("Gapless Dur", 7, "unused", False, False, False)
		self.coderef['coreaeGR'] = ("Gapless Resy", 7, "unused", False, False, False)
		self.coderef['coreaeGE'] = ("Gapless Enc Del", 5, "unused", False, False, False)
		self.coderef['coreasaa'] = ("Album Artist", 9, "albumartist", True, False, True)
		self.coderef['coreasgp'] = ("Gapless", 1, "unused", False, False, False)
		self.coderef['coreased'] = ("Extra Data", 2, "unused", False, False, False)
		self.coderef['coreashp'] = ("Has Been Played", 1, "unused", False, False, False)
		self.coderef['coreassn'] = ("Sort Name", 9, "unused", False, False, False)
		self.coderef['coreassa'] = ("Sort Artist", 9, "unused", False, False, False)
		self.coderef['coreassl'] = ("Sort Album Artist", 9, "unused", False, False, False)
		self.coderef['coreassu'] = ("Sort Album", 9, "unused", False, False, False)
		self.coderef['coreassc'] = ("Sort Composer", 9, "unused", False, False, False)
		self.coderef['coreasss'] = ("Sort Series Name", 9, "unused", False, False, False)
		self.coderef['coreasbk'] = ("Bookmarkable", 1, "unused", False, False, False)
		self.coderef['coreaspu'] = ("Podcast URL", 9, "unused", False, False, False)
		self.coderef['coreaeCR'] = ("iTunes Content Rating", 9, "unused", False, False, False)
		self.coderef['coreasai'] = ("iTunes Song Album ID", 7, "unused", False, False, False)
		self.coderef['coreasls'] = ("iTunes Song Long Size", 7, "unused", False, False, False)
		self.coderef['coreaeSE'] = ("iTunes Store Pers ID", 7, "unused", False, False, False)
		self.coderef['coreaeDV'] = ("iTunes DRM Versions", 5, "unused", False, False, False)
		self.coderef['coreaeDP'] = ("iTunes DRM Platform ID", 5, "unused", False, False, False)
		self.coderef['coreaeDR'] = ("iTunes DRM User ID", 7, "unused", False, False, False)
		self.coderef['coreaeND'] = ("iTunes Non DRM User ID", 7, "unused", False, False, False)
		self.coderef['coreaeK1'] = ("iTunes DRM Key1 ID", 7, "unused", False, False, False)
		self.coderef['coreaeK2'] = ("iTunes DRM Key2 ID", 7, "unused", False, False, False)
		self.coderef['coreaeXD'] = ("iTunes XID", 9, "unused", False, False, False)
		self.coderef['coreaeMk'] = ("iTunes Media Kind", 1, "unused", False, False, False)

	def startSocket(self):
		self.spsocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		self.spsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.spsocket.setblocking(0)
		self.spsocket.bind((self.dev.pluginProps["ipaddress"], int(self.dev.pluginProps["port"])))
		self.logger.debug("Socket Bound")

	def stopSocket(self):
		self.logger.debug("Closing socket for " + self.dev.name)
		self.spsocket.close
		self.spsocket = None

	def runDevConcurrentThread(self):
		# pretty much like runConcurrentThread, but at a device level, as opposed to a plugin level.
		self.logger.debug("Starting runDevConcurrentThread for " + self.dev.name)
		
		self.spsocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
		self.spsocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.spsocket.setblocking(0)
		self.spsocket.bind((self.dev.pluginProps["ipaddress"], int(self.dev.pluginProps["port"])))
		self.logger.debug("Socket Bound")
		self.bcastDict = {}
		self.bcastDict['none'] = ""
		self.bcastDict['devicename'] = ""
		self.bcastDict['clientip'] = ""
		self.bcastDict['useragent'] = ""
		self.bcastDict['srcremport'] = ""
		self.bcastDict['artoken'] = ""
		self.bcastDict['dacpid'] = ""
		self.clearBcastDict()
		
		bcastM = self.dev.pluginProps["shareMetadata"]
		if bcastM == True:
			self.logger.debug("Broadcast metadata for " + self.dev.name)
		
		playstate = False
		trackstart = round(time.time(), 0)
		tracklen = 120
		lastupdate = round(time.time(), 0)
		devid = str(self.dev.id)
		
		clearedTrackStates = [
			{'key':'added','value':""},
			{'key':'album','value':""},
			{'key':'albumartist','value':""},
			{'key':'artist','value':""},
			{'key':'bitrate','value':""},
			{'key':'bpm','value':""},
			{'key':'comment','value':""},
			{'key':'composer','value':""},
			{'key':'filedesc','value':""},
			{'key':'format','value':""},
			{'key':'genre','value':""},
			{'key':'modified','value':""},
			{'key':'name','value':""},
			{'key':'podcast','value':False},
			{'key':'samplerate','value':""},
			{'key':'tracknumber','value':""},
			{'key':'trackcount','value':""},
			{'key':'year','value':""}
		]
		
		try:
			bcastM = self.dev.pluginProps["shareMetadata"]
			reBcastTimes = False
			while not self.stopThread:
				try:
					time.sleep(sleeptime)
					data = self.spsocket.recv(1024)
					#self.logger.debug("Raw Data: " + str(data))
					sleeptime = 0.0001
					tag = data[:8]
					if tag in self.coderef:
						desc = self.coderef[tag][0]
						dtype = self.coderef[tag][1]
						dstate = self.coderef[tag][2]
						serverState = self.coderef[tag][3]
						storeLocal = self.coderef[tag][4]
						bcastData = self.coderef[tag][5]
						payload = data[8:]
						if dstate == "unused":
							pass
							#self.logger.debug(data[:8] + " found in self.coderef Dictionary, but currently unused in plugin. Discarding.")
						elif dtype == 0:
							# these types of metadata don't have any 'data'. Just their appearance indicates something happened.
							playstate = self.dataproc0(tag)
							self.dev.updateStateOnServer(dstate,playstate)					
							if bcastM == True:
								self.broadcastToSubscribers(u"ShairportData", (devid, dstate, playstate))
								reBcastTimes = True
						elif dtype == 1:
							restype1 = self.dataproc1(payload)
							if serverState == True:							
								self.dev.updateStateOnServer(dstate,restype1)
							if storeLocal == True:
								self.statedict[dstate] = restype1
							if bcastData == True:
								if bcastM == True:
									if sdtate in bcastDict:
										self.bcastDict[dstate] = restype1
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype1))
						elif dtype == 2:
							if data[:8] == "ssncmdst":
								self.logger.debug("New Metadata - Clear current states")
								self.dev.updateStatesOnServer(clearedTrackStates)
								self.clearBcastDict()
								#self.broadcastToSubscribers(u'ShairportData', (devid, "bcastDict", self.bcastDict))
								self.broadcastToSubscribers(u'ShairportData', (devid, "clearData", ""))
							if data[:8] == "ssncmden":
								self.broadcastToSubscribers(u'ShairportData', (devid, "bcastDict", self.bcastDict))
						elif dtype == 3:
							restype3 = self.dataproc3(payload)
							if serverState == True:							
								self.dev.updateStatesnServer(dstate,restype3)
							if storeLocal == True:
								self.statedict[dstate] = restype3
							if bcastData == True:
								if bcastM == True:
									if dstate in self.bcastDict:
										self.bcastDict[dstate] = restype3
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype3))
						elif dtype == 5:
							restype5 = self.dataproc5(payload)
							if serverState == True:							
								self.dev.updateStateOnServer(dstate,restype5)
							if storeLocal == True:
								self.statedict[dstate] = restype5
							if bcastData == True:
								if bcastM == True:
									if dstate in self.bcastDict:
										self.bcastDict[dstate] = restype5
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype5))
						elif dtype == 7:
							restype7 = self.dataproc7(payload)
							if serverState == True:							
								self.dev.updateStateOnServer(dstate,restype7)
							if storeLocal == True:
								self.statedict[dstate] = restype7
							if bcastData == True:
								if bcastM == True:
									if dstate in self.bcastDict:
										self.bcastDict[dstate] = restype7
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype7))
						elif dtype == 9:
							restype9 = self.dataproc9(payload)
							if len(payload) > 0:
								if serverState == True:	
									self.dev.updateStateOnServer(dstate,restype9)
								if storeLocal == True:
									self.statedict[dstate] = restype9
								if bcastData == True:
									if bcastM == True:
										if dstate in self.bcastDict:
											self.bcastDict[dstate] = restype9
										else:
											self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype9))
								### There's always one thing that needs special handling. Device ID doesn't seem to ever show up, so if
								### clientip does, grab the hostname and stick it in the variable
								if dstate == "clientip":
									hostname = socket.gethostname()
									if restype9 != socket.gethostbyname(hostname):
										hostname = socket.gethostbyaddr(restype9)[0]
									hostname = hostname.replace(".local","")
									hostname = hostname.replace("-"," ")
									hostname = ' '.join(word[0].upper() + word[1:] for word in hostname.split())
									hostname = hostname.replace("Iphone","iPhone")
									hostname = hostname.replace("Ipad","iPad")
									self.dev.updateStateOnServer('devicename', hostname)
									self.bcastDict['devicename'] = hostname
									#if bcastM == True:
									#	self.broadcastToSubscribers('ShairportData', (devid, 'devicename', hostname))
						elif dtype == 10:
							restype10 = self.dataproc10(payload)
							if serverState == True:							
								self.dev.updateStateOnServer(dstate,restype10)
							if storeLocal == True:
								self.statedict[dstate] = restype10
							if bcastData == True:
								if bcastM == True:
									if dstate in self.bcastDict:
										self.bcastDict[dstate] = restype10
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype10))
						elif dtype == 11:
							self.logger.debug(dself.ev.name + ": " + desc + ": Data Type 11 (version) -- Not Handled Yet:" + str(payload))
						elif dtype == 12:
							self.logger.debug(self.dev.name + ": " + desc + ": Data Type 12 (container) -- Not Handled Yet:" + str(payload))
						elif dtype == 13:
							restype13 = self.dataproc13(payload)
							if serverState == True:							
								self.dev.updateStateOnServer(dstate,restype13)
							if storeLocal == True:
								self.statedict[dstate] = restype13
							if bcastData == True:
								if bcastM == True:
									if dstate in self.bcastDict:
										self.bcastDict[dstate] = restype13
									else:
										self.broadcastToSubscribers(u'ShairportData', (devid, dstate, restype13))
						elif dtype == 14:
							trtimes = self.dataproc14(payload)
							tracklen = trtimes[0]
							progsec = trtimes[1]
							remsec = trtimes[2]
							trackstart = trtimes[3]
							if bcastM == True:
								timeDict = {}
								timeDict['tracklen'] = tracklen
								timeDict['progsec'] = progsec
								timeDict['nowplaying'] = playstate
								self.broadcastToSubscribers(u'ShairportData', (devid, "timeDict", timeDict))
							tftracklen = self.secToTime(tracklen)
							tfprogsec = self.secToTime(progsec)
							tfremsec = self.secToTime(remsec)
							tfAllTimes = [
								{'key':'tracklength', 'value':str(tftracklen)},
								{'key':'trackprogress', 'value':str(tfprogsec)},
								{'key':'trackremaining', 'value':str(tfremsec)}
							]
							self.dev.updateStatesOnServer(tfAllTimes)
							self.logger.debug(self.dev.name + " Track Length: " + tftracklen + " (" + str(tracklen) + " seconds)")
							self.logger.debug(self.dev.name + " Track Progress: " + tfprogsec + " (" + str(progsec) + " seconds)")
							self.logger.debug(self.dev.name + " Track Remaining: " + tfremsec + " (" + str(remsec) + " seconds)")
						else:
							self.logger.debug(self.dev.name + ": " + desc + ": Data Type unrecognized (" + str(dtype) + "): " + str(payload))
					else:
						#self.logger.debug(data[:8] + " NOT found in self.coderef dictionary.")
						pass
				except:
					sleeptime = 0.1
					if playstate == True:
						curstamp = round(time.time(), 0)
						if curstamp > lastupdate:
							progsec = int(curstamp - trackstart)
							#Grand Concerto can handle the stream level after receiving the initial time stamp.
							#But times always come AFTER the 'play' data, so on the first pass after play, resend the times.
							if bcastM == True:
								if reBcastTimes == True:
									timeDict = {}
									timeDict['tracklen'] = tracklen
									timeDict['progsec'] = progsec
									timeDict['nowplaying'] = playstate
									self.broadcastToSubscribers(u'ShairportData',(devid, 'timeDict',timeDict))
									reBcastTimes = False
							remsec = int(max(tracklen - progsec, 0))
							lastupdate = round(time.time(), 0)
							tfprogsec = self.secToTime(progsec)
							tfremsec = self.secToTime(remsec)
							tfAllTimes = [
								{'key':'trackprogress', 'value':tfprogsec},
								{'key':'trackremaining', 'value':tfremsec}
							]
							self.dev.updateStatesOnServer(tfAllTimes)
		except:
			pass
		finally:
			self.logger.debug("Closing socket for " + self.dev.name)
			self.spsocket.close
			self.spsocket = None

	def stopDevConcurrentThread(self):
		self.logger.debug("stopDevConcurrentThread called for " + self.dev.name)
		self.stopThread = True
		
	def clearBcastDict(self):
		# Specifically does NOT clear devicename or useragent or clientip
		self.bcastDict["added"] = ""
		self.bcastDict["album"] = ""
		self.bcastDict["albumartist"] = ""
		self.bcastDict["artist"] = ""
		self.bcastDict["bitrate"] = ""
		self.bcastDict["bpm"] = ""
		self.bcastDict["comment"] = ""
		self.bcastDict["composer"] = ""
		self.bcastDict["filedesc"] = ""
		self.bcastDict["format"] = ""
		self.bcastDict["genre"] = ""
		self.bcastDict["modified"] = ""
		self.bcastDict["name"] = ""
		self.bcastDict["podcast"] = ""
		self.bcastDict["samplerate"] = ""
		self.bcastDict["trackcount"] = ""
		self.bcastDict["tracknumber"] = ""
		self.bcastDict["year"] = ""

	def secToTime(self, seconds):
		hrs = divmod(seconds, 3600)[0]
		mins = divmod(divmod(seconds, 3600)[1], 60)[0]
		secs = divmod(seconds, 60)[1]
		if hrs == 0:
			tfh = ""
		else:
			tfh = str(hrs) + ":"
		
		if hrs == 0:
			tfm = str(mins) + ":"
		else:
			if mins < 10:
				tfm = "0" + str(mins) + ":"
			else:
				tfm = str(mins) + ":"
				
		if secs < 10:
			tfs = "0" + str(secs)
		else:
			tfs = str(secs)
				
		time = tfh + tfm + tfs
		return time
		
	def dataproc0(self, rawdata):
		if rawdata == "ssncpbeg":
			playstate = True
			stimg = "AvPlaying"
		elif rawdata == "ssncprsm":
			playstate = True
			stimg = "AvPlaying"
		elif rawdata == "ssncpfls":
			playstate = False
			stimg = "AvStopped"
		elif rawdata == "ssncpend":
			playstate = False
			stimg = "AvStopped"
		return playstate

	def dataproc1(self, rawdata):
		if len(rawdata) > 0:
			return True
		else:
			return False

	def dataproc2(self, rawdata):
		return rawdata

	def dataproc3(self, rawdata):
		res = struct.unpack('!h', rawdata)
		return str(res[0])

	def dataproc5(self, rawdata):
		res = struct.unpack('!l', rawdata)
		return str(res[0])

	def dataproc7(self, rawdata):
		res = struct.unpack('!q', rawdata)
		return str(res[0])

	def dataproc9(self, rawdata):
		return rawdata

	def dataproc10(self, rawdata):
		rawdate = struct.unpack('!i', rawdata)
		repdatetime = datetime.datetime.fromtimestamp(rawdate[0])
		repdate = datetime.date(repdatetime.year, repdatetime.month, repdatetime.day)
		return str(repdate)

	def dataproc11(self, rawdata):
		return rawdata

	def dataproc12(self, rawdata):
		return rawdata

	def dataproc13(self, rawdata):
		timesec = int(rawdata) / 44100
		return timesec

	def dataproc14(self, rawdata):
		rawprog = rawdata.split("/")
		tracklen = self.dataproc13(int(rawprog[2]) - int(rawprog[0]))
		tprogsec = self.dataproc13(int(rawprog[1]) - int(rawprog[0]))
		tremsec = self.dataproc13(int(rawprog[2]) - int(rawprog[1]))
		trackstart = round(time.time(), 0) - tprogsec
		return tracklen, tprogsec, tremsec, trackstart
		